
#include "dlgwin32.h"
#include "debug.h"
#include "resource.h"

#include <process.h>
#include <assert.h>



///////////////////////////////////////////////////////////////////////////////


int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow ) {

 	::DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINWND), NULL, DialogProc );

	return 0;
}


///////////////////////////////////////////////////////////////////////////////


BOOL CALLBACK DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {

	switch(uMsg) {
	HANDLE_DLG_MSG (hwnd, WM_INITDIALOG, OnInitDialog);
	HANDLE_DLG_MSG (hwnd, WM_COMMAND, OnCommand);
	}

	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////


BOOL OnInitDialog(HWND hwnd, HWND hWndFocus, LPARAM lParam) {

	SetDlgItemText ( hwnd, IDC_PATH, TEXT("C:\\foo.dat") );
	return TRUE;

}


///////////////////////////////////////////////////////////////////////////////


void OnCommand(HWND hwnd, int nID, HWND hWndCtl, UINT codeNotify) {

	switch(nID) {
	case IDC_READ: {

		TCHAR szPathFrom [ MAX_PATH ];
		Edit_GetText(GetDlgItem(hwnd, IDC_PATH), szPathFrom, sizeof (szPathFrom) );

		TCHAR szPathTo [MAX_PATH];
		if (!GetTempFileName ( TEXT("C:\\temp"), TEXT("adf"), 0, szPathTo ) ) {
			::MessageBox ( hwnd, TEXT("GetTempFileName failed."), TEXT("Error"), MB_OK | MB_ICONERROR );
			break;
		}

		SetDlgItemText ( hwnd, IDC_TO_FILE, szPathTo );

		BeginAsyncRead ( hwnd, szPathFrom, szPathTo );

		break;
	}
	case IDCANCEL:
		EndDialog(hwnd, 0);
		break;

	}
}


///////////////////////////////////////////////////////////////////////////////


VOID BeginAsyncRead ( HWND hwnd, LPTSTR pszPathFrom, LPTSTR pszPathTo ) {

	//
	// Open the file (FROM)
	//
	
	HANDLE hFile = CreateFile ( 
		pszPathFrom, 
		GENERIC_READ,
		NULL,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		NULL);

	if ( INVALID_HANDLE_VALUE == hFile ) {
		::MessageBox ( hwnd, TEXT("CreateFile failed - hFile."), TEXT("Error"), MB_OK | MB_ICONERROR);
		return;
	}

	//
	// Open the file (TO)
	//

	HANDLE hSaveFile = CreateFile (
		pszPathTo,
		GENERIC_WRITE,
		NULL,
		NULL,
		OPEN_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,
		NULL);

	if ( INVALID_HANDLE_VALUE == hSaveFile ) {
		::MessageBox ( hwnd, TEXT("CreateFile failed - hSaveFile."), TEXT("Error"), MB_OK | MB_ICONERROR);
		CloseHandle ( hFile );
		return;
	}

	
	//
	// Create a CONTEXT
	//
	
	PREAD_CONTEXT pContext = (PREAD_CONTEXT) malloc ( sizeof (READ_CONTEXT) );
	DWORD cbFileSize = GetFileSize ( hFile, NULL );
	
	if ( !pContext || INVALID_FILE_SIZE == cbFileSize ) {
		::MessageBox ( hwnd, TEXT("Failed to allocate the memory or get the file size."), TEXT("Error"), MB_OK | MB_ICONERROR);
		CloseHandle ( hFile );
		CloseHandle ( hSaveFile );
		return;
	}
	ZeroMemory ( pContext , sizeof (READ_CONTEXT) );

	pContext->hFile = hFile;
	pContext->cbTotalBytesRead = 0;
	pContext->cbFileSize = cbFileSize;
	pContext->ol.Offset = 0;
	pContext->hSaveFile = hSaveFile;
	
	//
	// Begin a new thread
	//

	pContext->hThread = (HANDLE) _beginthreadex(
		NULL, 
		NULL, 
		WorkerThreadFunc, 
		(void*) pContext, 
		NULL, 
		&pContext->uThreadId);

	if ( !pContext->hThread ) {
		
		::MessageBox (hwnd, TEXT("Failed to begin a new thread."), TEXT("Error"), MB_OK | MB_ICONERROR );

		CleanupContext ( pContext );

	}

}


///////////////////////////////////////////////////////////////////////////////


VOID CleanupContext ( PREAD_CONTEXT pContext ) {

	if ( !pContext ) {
		return;
	}

	if ( pContext->hFile ) {
		CloseHandle ( pContext->hFile );
		pContext->hFile = NULL;
	}
	
	if ( pContext->hSaveFile ) {
		CloseHandle ( pContext->hSaveFile );
		pContext->hSaveFile = NULL;
	}

	free ( pContext );

}



///////////////////////////////////////////////////////////////////////////////


VOID CALLBACK FileIOCompletionRoutine( 
	DWORD dwErrorCode, 
	DWORD dwNumberOfBytesTransfered, 
	LPOVERLAPPED lpOverlapped ) {

	//
	// Tracing
	//
	
	DebugPrint ( TEXT("FileIOCompletionRoutine:\nErrorCode: %u\nBytesTransferred: %u\n"), dwErrorCode, dwNumberOfBytesTransfered );

	//
	// Retreive PREAD_CONTEXT using the address of OVERLAPPED.
	//
	
	PREAD_CONTEXT pContext = (PREAD_CONTEXT) CONTAINING_RECORD (lpOverlapped, READ_CONTEXT, ol);

	//
	// Update the context
	//

	pContext->cbTotalBytesRead += dwNumberOfBytesTransfered;
	pContext->ol.Offset += dwNumberOfBytesTransfered;

	//
	// Use the buffer
	//
	DWORD cbNumberOfBytesWritten = 0;
	BOOL bWrite;
	bWrite = WriteFile ( 
		pContext->hSaveFile, 
		pContext->pBuffer,
		dwNumberOfBytesTransfered,
		&cbNumberOfBytesWritten,
		NULL );

	if ( !bWrite ) {
		DebugPrint ( TEXT("WriteFile failed %u\n"), GetLastError() );
		CleanupContext ( pContext );
		return;
	}

	// DebugHexDump ( dwNumberOfBytesTransfered, (PBYTE) pContext->pBuffer );
	ZeroMemory ( pContext->pBuffer, BUFFER_SIZE );

	//
	// Need to read again?
	//

	if ( pContext->cbFileSize == pContext->cbTotalBytesRead ) {

		//
		// Read the file completely.
		//
		
		DebugPrint ( TEXT("OK!!\n") );

		CleanupContext ( pContext );
		
		//
		// Exit this thread.
		//

		DebugPrint ( TEXT("WorkerThreadFunc %08x End - Calling ExitThread(0) \n"), GetCurrentThreadId() );
		
		ExitThread (0);
		return;
	}
		
	//
	// Call ReadFile again.
	//

	BOOL bRet = ReadFileEx (
		pContext->hFile,
		pContext->pBuffer,
		min (BUFFER_SIZE, (pContext->cbFileSize - pContext->cbTotalBytesRead)),
		&pContext->ol,
		FileIOCompletionRoutine );

	if ( !bRet || ( bRet && ERROR_SUCCESS != GetLastError() ) ) {
		DebugPrint ( TEXT("ReadFileEx failed %u\n"), GetLastError() );

		CleanupContext ( pContext );
	}
	
}


///////////////////////////////////////////////////////////////////////////////


unsigned int __stdcall WorkerThreadFunc (PVOID pv) {

	DebugPrint ( TEXT("WorkerThreadFunc %08x Start\n"), GetCurrentThreadId() );

	PREAD_CONTEXT pContext = (PREAD_CONTEXT) pv;

	BOOL bRet = ReadFileEx (
		pContext->hFile,
		pContext->pBuffer,
		BUFFER_SIZE,
		&pContext->ol,
		FileIOCompletionRoutine );
		
	if ( !bRet  || ( bRet && ERROR_SUCCESS != GetLastError() ) ) {

		TCHAR szErrorMessage [256];
		wsprintf ( szErrorMessage, TEXT("ReadFileEx failed (%u)"), GetLastError() );
		::MessageBox ( NULL, szErrorMessage, TEXT("Error"), MB_OK | MB_ICONERROR );

		CleanupContext ( pContext );

		return 0;
	}

	while ( SleepEx ( INFINITE, TRUE ) ) {
	}

	DebugPrint ( TEXT("WorkerThreadFunc %08x End\n"), GetCurrentThreadId() );

	return 0;
}

